#' Estimate the regressor coefficients using smoothing spline in varying-coefficient models.
#' 
#' @param Y The longitudinal observation matrix of size \code{m} (subjects) by \code{n}(time points)
#' @param X The regressor matrix of size \code{m}(subjects) by \code{n}(timepoints) by \code{k}(dimension)
#' @param tvec The timepoint vector of length \code{n}
#' @param lambda The smoothing parameter. If NULL then use GCV optimized smoothing parameter
#' @param Wvec The weight vector
#' @return A list of fitted values, residuals, smoothing parameter, estimated varying coefficient
#' 
#' @export
NPfit<-function(Y,X,tvec,lambda=NULL,Wvec=NULL){
  library("psych")
  m=dim(Y)[1];n=dim(Y)[2];k=dim(X)[3]
  tscale<-(tvec-min(tvec))/(max(tvec)-min(tvec)); 
  grid_start<-(10^(-8))*rep(1,times=k)
  Qmat0=array() 
  for (ii in seq(1,n,1)){  
    for (jj in seq(1,n,1)){  
      qtilde<-function(s){(tscale[ii]-s)*(tscale[jj]-s)}
      Qmat0[(ii-1)*n+jj]<-integrate(qtilde, lower = 0, upper = min(tscale[ii],tscale[jj]))
    } 
  } 
  Qmat1<-matrix(as.numeric(Qmat0),ncol=n, byrow=TRUE) 
  lstT=list() 
  ttilde=rbind(rep(1,length=n),tscale)  
  for (j in seq(1,n,1)){  
    lstT[[j]]<-kronecker(t(ttilde[,j]),diag(k))  
  }
  Ttilde=do.call(rbind,lstT) 
  lstH=list();#lstY=list();   
  for (i0 in seq(1,n,1)){ 
    vh=matrix(0,m*n,k); 
    vh[seq((i0-1)*m+1,i0*m),]=X[,i0,];  
    lstH[[i0]]=vh;   
  }
  H<-do.call(cbind,lstH);
  ycol=array(Y)
  V=H%*%Ttilde; 
  if(is.null(Wvec)){Wvec<-rep(1,times=n*m)} 
  W<-diag(array(matrix(Wvec,ncol=n,byrow=TRUE)),n*m)
  if(is.null(lambda)){
    ## lambda-related variables ##
    GCVscore<-function(lambda){
      Afun<-function(lambda){  
        Mfun<-function(lambda){
          Qmatfun<-function(lambda){
             kronecker(Qmat1,diag(lambda^(-1),k))
          } 
          H%*%Qmatfun(lambda)%*%t(H)%*%W+diag(n*m)
        }
        Qmatfun<-function(lambda){
          kronecker(Qmat1,diag(lambda^(-1),k))
        }
        diag(n*m)-(solve(Mfun(lambda))-solve(Mfun(lambda))%*%V%*%solve(t(V)%*%W%*%solve(Mfun(lambda))%*%V)%*%t(V)%*%W%*%solve(Mfun(lambda)));
      }
      t(array(Afun(lambda)%*%ycol)-ycol)%*%(array(Afun(lambda)%*%ycol)-ycol)/(1-tr(Afun(lambda))/(m*n))^2
    }
    lambdaest<-constrOptim(grid_start, GCVscore, grad=NULL,ui=diag(k),ci=rep(0,times=k))
    lambda<-lambdaest$par
  }
  Qmat<-kronecker(Qmat1,diag(lambda^(-1),k)) 
  M=H%*%Qmat%*%t(H)%*%W+diag(n*m); 
  C=Qmat%*%t(H)%*%W%*%solve(M)-(Qmat%*%t(H)%*%W%*%solve(M)%*%V-Ttilde)%*%solve(t(V)%*%W%*%solve(M)%*%V)%*%t(V)%*%W%*%solve(M);
  A=diag(n*m)-(solve(M)-solve(M)%*%V%*%solve(t(V)%*%W%*%solve(M)%*%V)%*%t(V)%*%W%*%solve(M));
  vecbetahat=matrix(C%*%ycol,nrow=k,ncol=n); 
  fity=A%*%ycol; 
  ##--------------------output---------------------##
  output.df=data.frame('time'=rep(tvec,times=m),
                       'group'=rep(c(1:m),each=n),
                       'Y'=array(t(Y)),
                       'fitted'=array(matrix(fity,ncol=m,byrow=TRUE)),
                       'resid'=array(matrix(ycol-array(fity),ncol=m,byrow=TRUE)))
  Ymat.fit<-matrix(fity,ncol=n,byrow=TRUE) 
  rmat<-matrix(ycol-array(fity),ncol=n,byrow=TRUE)
  returnlist=list('Ymat.fit'=Ymat.fit,
                  'rmat'=rmat,
                  'beta'=vecbetahat,
                  'sp'=lambda,
                  'fitted'=output.df$fitted,
                  'residuals'=output.df$resid,
                  'output.df'=output.df)
  return(returnlist)
}


