#' Implement stability constraint to varying coefficient model using smoothing spline
#' 
#' @param Ymat The longitudinal observation matrix of size \code{m} (subjects) by \code{n}(time points)
#' @param Xmat The regressor matrix of size \code{m}(subjects) by \code{n}(timepoints) by \code{k}(dimension)
#' @param tvec The timepoint vector of length \code{n}
#' @param lambda The smoothing parameter.
#' @param h The bandwidth parameter. \code{h=1} indicates the bandwidth is 1 standard deviation. 
#' @param alpha The significance level
#' @param max.iter The maximal number of iteration.
#' @return A list with components
#'    \itemize{
#'      \item beta - estimated varying coefficient after stability constraint implementation
#'      \item bound - The upper and lower bound 
#'      \item convScore - The convergence score
#'      \item conv.time - The timepoint from which stability constraint applies
#'      \item stop criteria - The stopping criteria
#'      \item residuals - The difference between observation and fitted values
#'      \item fitted - The fitted values after implementing stability constraint
#'   }
#' @export
ConstrS<-function(Ymat,Xmat,tvec,lambda,h=NULL,alpha=NULL,max.iter=NULL){
  ##-------------- initialize --------------------------##
  library('DescTools')
  source("GCVNPfit.R")
  m=dim(Ymat)[1]; n=dim(Ymat)[2];k=dim(Xmat)[3];
  if(is.null(alpha)){alpha=0.005}
  if(is.null(h)){h=1}
  if(is.null(max.iter)){max.iter=100}
  tscore<-qt(1-alpha/2, df=n-1) 
  k0<-0
  temp.df<-data.frame('time'=rep(tvec,times=m),
                      'Y'=array(t(Ymat)),
                      'group'=rep(c(1:m),each=n)
                        )
  beta.list<-list(); cv.vec<-array();
  ##--------- estimation of sigmahat -----------##
  siglist=list();
  for (i in c(1:m)){
    rc=array();
    for (j in c(2:n)){
      rc[j-1]<-temp.df$Y[temp.df$group==i&temp.df$time==tvec[j]]-temp.df$Y[temp.df$group==i&temp.df$time==tvec[j-1]]
    }
    siglist[[i]]<-median(abs(rc))
  }
  sighat<-1.48/sqrt(2)*do.call(median,siglist)
  ##--------- convergence range ----------------##
  rca=list();
  for (i in c(1:m)){
    rc=array();
    for (j in c(1:n)){
      rc[j]<-mean(temp.df$Y[temp.df$group==i&temp.df$time>=tvec[j]])-temp.df$Y[temp.df$group==i&temp.df$time==tvec[j]]
    }
    bound=which(rc<=h*sighat&rc>=-1*h*sighat)
    rca[[i]]<-c(min(bound),max(bound))
  }
  ta<-max(do.call(rbind,rca)[,1]); #tb<-max(do.call(rbind,rca)[,2]);
    ##---------------------------------------------##
  #temp.df$comp.np<-return.list$comp_np
  temp.df$lower<-temp.df$Y-tscore*sighat/sqrt(n); temp.df$upper<-temp.df$Y+tscore*sighat/sqrt(n);
  mdl.update= NPfit(Ymat,Xmat,tvec,lambda=lambda,Wvec=NULL)
  temp.df$r<-mdl.update$residuals
  temp.df$fitted<-mdl.update$fitted
  phi0<-length(temp.df$fitted[temp.df$fitted<temp.df$lower|temp.df$fitted>temp.df$upper])
  phi<-phi0
  bound.df<-data.frame("lower"=temp.df$lower,
                       "upper"=temp.df$upper)
  ##---------------- iterative --------------------------##
  while (phi<=phi0&k0<=max.iter){
    k0<-k0+1
    ##-----------check convergence performance-----##
        beta.fit<-data.frame('time'=tvec,'betahat'=array(mdl.update$beta))
        betafun<-splinefun(beta.fit$time,beta.fit$beta,method='natural')
        ls(envir = environment(betafun))
        splinecoef <- get("z", envir = environment(betafun))
        c1<-curve(abs(betafun(x,deriv=1)),tvec[ta],tvec[n]);
        cv.vec[k0]<-AUC(c1$x,c1$y, method='spline')
        beta.list[[k0]]<-array(mdl.update$beta)
    ##----------------------------------------------##
    r0<-min(abs(temp.df$r[temp.df$time>=tvec[ta]]))
    temp.df$W<-(temp.df$time<tvec[ta])+(1-(temp.df$time<tvec[ta]))*r0/abs(temp.df$r)
    mdl.update= NPfit(Ymat,Xmat,tvec,lambda=lambda,Wvec=temp.df$W)
    temp.df$r<-mdl.update$residuals
    temp.df$fitted<-mdl.update$fitted
    phi<-length(temp.df$fitted[temp.df$fitted<temp.df$lower|temp.df$fitted>temp.df$upper])
 }
    flag<-c("exceed accuracy tolerance","exceed maximum iteration")[1+(k0>max.iter)]
  ##--------------------- return ------------------------##
  idx<-which.min(cv.vec)
  fitted.list<-list()
  for (i in c(1:m)){
    fitted.list[[i]]<-Xmat[i,,]*beta.list[[idx]]
  }
  fitmat<-do.call(rbind,fitted.list)
  fitted<-as.vector(t(fitmat))
  resid<-temp.df$Y-fitted
  
    returnlist<-list(beta.list[[idx]],
                     bound.df,
                     min(cv.vec),
                     tvec[ta],
                     flag,
                     resid,
                     fitted)
    names(returnlist)<-c("beta",
                         "bound",
                         "convScore",
                         "conv.time",
                         "stop criteria",
                         "residuals",
                         "fitted")
  return (returnlist)
    
}
  #======== compute the derivative function===============#  
  #g <- betafun
  #environment(g) <- environment(betafun)#new.env(parent=baseenv())
  #z <- get("z", envir=environment(betafun)) 
  #z$y <- z$b
  #z$b <- 2*z$c
  #z$c <- 3*z$d
  #z$d <- rep(0, z$n)
  #assign("z", z, envir=environment(g))
  #curve(abs(g(x)), 1, 10, col = "green", lwd = 1.5)
  
  ##--------plot -----------------##
  #pred.beta<-spline(beta.fit$time,beta.fit$beta,method='natural',n=500)
  #plot(pred.beta$x,pred.beta$y,type='l',lty=3,'col'='red')
  #abline(pred.beta0$x,pred.beta0$y,type='l',lty=3,col='green')
 
