#' Estimate semi-parametric model 
#' Estimate unknown parameters in semiparametric model using least square for parametric part and smoothing spline in nonparametric part.
#' 
#' @param fmr The parametric expression n the model 
#' @param data The dataframe including observation \code{Y},regressor \code{X} and time points  \code{time}
#' @param Xmat Optional. The matrix of regressor with size \code{m}(subjects) by \code{n}(timepoints) by \code{k}(dimension)
#' @param theta_start Initial value of parameters in parametric expression 
#' @param max.iter The maximum number of iteration (200 by defualt)
#' @param tol The tolerance of convergence (0.001 by default for each parameter)
#' @param lambda The smoothing parameter
#' @return A list with components
#'   \itemize{
#'      \item thea - the estimated parameter in the parametrc part 
#'      \item beta - the estimated parameters in the nonparametric part
#'      \item iter.num - the number of iteration processed
#'      \item sp - the smoothing parameter 
#'      \item p.model - the fitted model of parametric part 
#'      \item comp.np - the fitted values from nonparametric part
#'      \item comp.p - the fitted values from parametric part 
#'      \item fitted - the fitted values of the model 
#'      \item residuals - the difference between fitted and observed values 
#'      \item theta.conv - the vector of change in parameters per iteration
#'      \item output.df - the dataframe contains original data, fitted values and residuals
#'   }
#'      
#' @export
semi<-function(fmr,data,Xmat=NULL,theta_start,max.iter=NULL,tol=NULL,lambda=NULL){
  ##--------- set estimation parameters ------------------##
  source("GCVNPfit.R")
  tvec<-unique(data$time)
  temp.df<-data
  temp.df$Y_r<-data$Y
  theta_new<-do.call(cbind,theta_start)
  i.count=0
  td=NULL
  if(is.null(tol)){
    tol=length(theta_start)*0.001
  }
  theta_diff<-tol+1
  
  if(is.null(max.iter)){
    max.iter=200
  }
  
  if(is.null(Xmat)){
    k=1;m=length(unique(data$X)); n=length(tvec);
    Xmat=array(rep(unique(data$X),times=n*k),c(m,n,k))
  } else{
    m=dim(Xmat)[1];n=dim(Xmat)[2];k=dim(Xmat)[3];
  }
  
  ##------------------------Iterative fitting -----------------------------##
  while (theta_diff>tol&i.count<max.iter){
    ##--- step1: estimate parametric function ---#
    theta_old<-theta_new
    m1.mdl<-nls(fmr,data=temp.df,start=theta_start)
    theta_new<-summary(m1.mdl)$coefficient[,1]
    theta_diff<-sum(abs(theta_new-theta_old))
    
    temp.df$Y_p<-predict(m1.mdl, temp.df$time)
    #alternatively,temp.df$resid_p<-summary(step1.mdl)$residuals
    temp.df$resid_p<-temp.df$Y-temp.df$Y_p
    ##--- step2: estimate nonparametric function ---#  
    Ymat=matrix(temp.df$resid_p,ncol=n,byrow=TRUE)  # data structure
    m2.mdl= NPfit(Ymat,Xmat,tvec,lambda=lambda,W=NULL)
    temp.df$Y_np<-m2.mdl$fitted
    #alternatively, temp.df$resid<-temp.df$Y-temp.df$Y_np-temp.df$Y_p
    temp.df$resid_np<-temp.df$Y-temp.df$Y_np
    #alternatively, temp.df$resid_np<-temp.df$resid_p-temp.df$Y_p
    temp.df$Y_r<-temp.df$resid_np
      #temp.df$X*rep(m2.mdl$beta,times=m) 
    
    i.count<-i.count+1  
    td<-c(td,theta_diff) 
    
  }
  
  #theta<-theta_new
  #beta<-m2.mdl$beta
  #sp<-m2.mdl$sp
  #comp.np<-temp.df$Y_np
  #comp.p<-temp.df$Y_p
  #fitted<-temp.df$Y_np+temp.df$Y_p
  #residuals<-temp.df$Y-fitted
  #p.mdl<-m1.mdl
  
  output.df<-data.frame(data,
                        'fitted'=temp.df$Y_np+temp.df$Y_p,
                        'residuals'=temp.df$Y-(temp.df$Y_np+temp.df$Y_p))
  
  return.list<-list(theta_new,
                    m2.mdl$beta,
                    i.count,
                    m2.mdl$sp,
                    m1.mdl,
                    temp.df$Y_np,
                    temp.df$Y_p,
                    temp.df$Y_np+temp.df$Y_p,
                    temp.df$Y-(temp.df$Y_np+temp.df$Y_p),
                    td,
                    output.df)
    
  names(return.list)<-c('theta',
                        'beta',
                        'iter.num',
                        'sp',
                        'p.model',
                        'comp.np',
                        'comp.p',
                        'fitted',
                        'residuals',
                        'theta.conv',
                        'output.df')
  return(return.list)
}
  
  
  
  
  
  
  
  


